.RecordAction-label {
    font-size: 15px;
    font-weight: bold;
}

.area-selection {
    background-color: rgba(255, 0, 0, 0.5);
    border: 2px solid rgba(255, 0, 0, 1);
}

.area-recording {
    border-style: dashed;
    border: 2px rgba(255, 0, 0, 1);
}

.alert-msg {
    font-size: 40px;
    font-weight: bold;
    color: #ffffff;
    background-color: rgba(10, 10, 10, 0.7);
    border-radius: 5px;
    padding: .5em;
}

.area-resolution {
    font-size: 20px;
    font-weight: bold;
    color: #ffffff;
    background-color: rgba(0, 0, 0, 0.0);
}

.time-label {
    font-size: 14px;
    padding-right: 5px;
    color: #ff0000;
}
